"""
=============================================================
                           CONTACT
=============================================================

  Add me on discord : 		@kalgon57

=============================================================
                           DONATE
=============================================================

  Appreciate my program ? Does it help you manage your cryptocurrencies ?
  If so, I thank you for considering making a donation. No matter the amount, it will be greatly appreciated.

  Bitcoin				          bc1qu7lfaxmu54vg4mr05t69ruxc7jp88t0xs4qp73

  Ethereum and other EVMs		  0x46969a640cbe700a49c72228886c3ea646e9b618

  MultiversX			          erd1972hx8rjh7hhc5np890qp6lzl99yj8vg3zvw268ehm6tla4tt3ksq7hm5c

=============================================================
                      LICENSE AGREEMENT
=============================================================

  Author: kalgon57
  This program is provided under the terms of the "k57-portfolio" License Agreement.
  By using this software, you agree to the following terms and conditions:

  **k57-portfolio License Agreement**

    1. **Usage**:
       - This program is provided for private use only. It may not be used for commercial purposes.
    2. **Modifications**:
       - You are permitted to modify the program for private use only. Any modified versions may not be used for commercial purposes.
    3. **Distribution**:
       - You may distribute the original version of the program freely, provided that all copyright notices and this license agreement are included.
    4. **No Warranty**:
       - This program is provided "as is" without warranty of any kind, express or implied. In no event shall the authors or copyright holders be liable for any claim, damages, or other liability, whether in an action of contract, tort, or otherwise, arising from, out of, or in connection with the program or the use or other dealings in the program.
    5. **Jurisdiction**:
       - This license agreement shall be governed by and construed in accordance with the laws of Switzerland.

    For questions or concerns regarding this license agreement, please contact @kalgon57 on discord.

  **Credits**
    - Python 3.11
    - Flask 3.0.0
    - Requests 2.31.0
    - jQuery 3.6.0
    - Chart.js 4.4.1
    - PyCharm 2023.1
    - flaticon.com
    - freepik.com
    - cointelegraph.com
    - coinmarketcap.com
    - coingecko.com
    - cryptolasereyes.com

"""

import logging
from flask import jsonify, make_response, render_template, Response, Flask, redirect, request, send_file
import pathlib
import json
import sqlite3
import datetime
import requests
from random import shuffle
from types import SimpleNamespace
from typing import Optional, Union, List, Dict, Tuple, Any, TypeAlias

DEBUG = False

SQLR:TypeAlias = Tuple[str,Tuple]
SQLRequests: TypeAlias = Union[List[SQLR],SQLR]
DTFORMAT = '%Y-%m-%d %H:%M'
COLORS = ['#9552EA', '#93F03B', '#F54F52', '#FFA32F', '#378AFF', '#FFEC21', '#FB33F3', '#2C9D3E', '#7CDDDD']
COLORS_DEBT = ['#222222', '#444444', '#666666', '#888888', '#333333', '#555555','#777777', '#999999', '#222222']
CMC_VIEWER = 'https://coinmarketcap.com/currencies/'
CG_VIEWER = 'https://www.coingecko.com/en/coins/'

def get_logger() -> logging.Logger:
  """
  Initialize the logger used for the entire application.
  Messages are recorded in "portfolio.log".
  """
  logger = logging.getLogger(__name__)
  logger.setLevel(logging.INFO)
  file_handler = logging.FileHandler('portfolio.log')
  formatter = logging.Formatter(fmt='{asctime} ; {levelname:>8s} ; {message}', style='{', datefmt=DTFORMAT)
  file_handler.setFormatter(formatter)
  logger.addHandler(file_handler)
  return logger

# Formatting strings for displaying purpose, for python and jinja
def fixed(apr:float,d:int) -> str:
  """
  Exactly d decimal(s).
  """
  if apr is None:
    return ''
  return f'{apr:.{d}f}'
def format_price_round(price:float) -> str:
  """
  No decimal.
  Spacing each 3 digits.
  """
  return f'{price:,.0f}'.replace(',',' ')
def format_price(price:float) -> str:
  """
    0 decimals for "high" prices.
    2 decimals for "normal" prices.
    3 significant digits for "low" prices.
    Spacing each 3 digits.
  """
  if price>100:
    return f'{price:,.0f}'.replace(',',' ')
  if price>=1:
    return f'{price:.2f}'
  if price==0:
    return '0'
  s = f'{price:.12f}'
  for i in range(2,len(s)):
    if s[i]!='0':
      return s[:i+3]
  return s
def format_datetime(dt:datetime.datetime) -> str:
  """
  Standard format without second, or "unknown".
  """
  return dt.strftime(DTFORMAT) if dt is not None else 'unknown'
def format_time_elapsed(dt:datetime.datetime) -> str:
  """
  Returns the time elapsed until now in a natural human language.
  """
  delta = datetime.datetime.now()-dt
  delta = delta.total_seconds()
  if delta<3600:
    return f'{round(delta/60)} min ago'
  if delta<86400:
    return f'{round(delta/3600)} hours ago'
  return f'{round(delta/86400)} days ago'
def format_time_remaining(dt:datetime.datetime) -> str:
  """
  Returns the time remaining from now in a natural human language.
  """
  if dt is None:
    return 'undefined'
  now = datetime.datetime.now()
  if now>=dt:
    return 'now <img src="/static/img/now.png" title="Now" alt="NOW">'
  delta = dt-now
  delta = delta.total_seconds()
  if delta < 3600:
    return f'in {round(delta/60)} min'
  if delta < 86400:
    return f'in {round(delta/3600)}hours'
  return f'in {round(delta/86400)} days'
# Managing 'None' values
def if_none(var:Any, default) -> Optional[Any]:
  """
  Replaces None by default.
  """
  return var if var is not None else default
def str_or_none(s:str) -> Optional[str]:
  """
  Replaces empty string by None.
  """
  return s if s!='' else None

class Portfolio:
  def __init__(self):
    self.logger = get_logger()
    self.config = None
    self.load_config()
    self.last_update = None
    self.update_delay = None
    if self.config['price_update_on_server_start']:
      self.update_prices()
    self.data = None
    self.load_data()
    self.server = Flask(__name__)
    self.define_rules()
  def run(self) -> None:
    """
    Runs the Flask web server.
    """
    self.server.run(host='127.0.0.1', port=self.config['port'], debug=DEBUG)
  def load_config(self) -> None:
    """
    Load the "config.json" file and check its integrity (variables and their types).
    Raises an exception if something goes wrong.
    """
    try:
      json_file = pathlib.Path(__file__).parent / 'config.json'
      json_content = json_file.read_text()
      self.config = json.loads(json_content)
    except FileNotFoundError:
      self.logger.error('"config.json" was not found.')
      raise Exception
    except json.JSONDecodeError:
      self.logger.error('"config.json" contains syntax error(s).')
      raise Exception
    config_struct = {
      'port':int,
      'database_filename':str,
      'api_provider':str,
      'api_key':str,
      'price_update_on_server_start':bool,
      'percent_small_variation':int,
      'percent_big_variation':int,
      'token_viewer': str
    }
    if not all(key in self.config for key in config_struct.keys()):
      self.logger.error('"config.json" does not define required variables.')
      raise Exception
    for k,v in config_struct.items():
      if not (isinstance(self.config[k], v)):
        self.logger.error(f'"{k}" variable in "config.json" should be an instance of "{v.__name__}".')
        raise Exception
  def db_execute(self, request:SQLRequests) -> Optional[List]:
    """
    Executes one or more SQL request(s).
    Rollbacks and returns None if something goes wrong.
    Returns a list of results otherwise.
    """
    last_request = ('',None)
    try:
      with sqlite3.connect(self.config['database_filename']) as db_conn:
        db_cursor = db_conn.cursor()
        db_cursor.execute('PRAGMA foreign_keys = ON;')
        if isinstance(request,tuple):
          last_request = request
          db_cursor.execute(*request)
        else:
          for req in request:
            last_request = req
            db_cursor.execute(*req)
        db_conn.commit()
        result = db_cursor.fetchall()
      return result
    except sqlite3.Error as e:
      self.logger.warning(f'Database: execution failed: {last_request}. {e}')
      return None
  def update_prices(self) -> bool:
    """
    Gets prices from the configured API (CoinMarketCap or CoinGecko) and updates the database.
    Returns False if something goes wrong, True otherwise.
    CoinGecko API is not implemented yet.
    Each API case should populate the "dic" dictionary as follows:
      - Keys are the slugs.
      - Values are SimpleNameSpaces with "ide" attribute (from database) and "price" attribute (from API).
    """
    dic = {}
    if self.config['api_provider']=='cmc':
      array = if_none(self.db_execute(('SELECT Id, SlugCMC FROM Token;',())),[])
      for row in array:
        if row[1]:
          c = SimpleNamespace()
          c.ide = row[0]
          c.price = None
          dic[row[1]] = c
      if len(dic)==0:
        self.last_update = datetime.datetime.now()
        return True
      url = 'https://pro-api.coinmarketcap.com/v2/cryptocurrency/quotes/latest'
      parameters = {'slug': ','.join(dic.keys())}
      headers = {'Accepts': 'application/json', 'X-CMC_PRO_API_KEY': self.config['api_key']}
      session = requests.Session()
      session.headers.update(headers)
      try:
        response = session.get(url, params=parameters)
        cryptos = json.loads(response.text)
      except:
        self.logger.warning('Unable to update prices: API request failed.')
        return False
      if cryptos['status']['error_code']:
        self.logger.warning(f'Unable to update prices: {cryptos["status"]["error_message"]}.')
        return False
      for crypto in cryptos['data'].values():
        dic[crypto['slug']].price = crypto['quote']['USD']['price']
    elif self.config['api_provider'] == 'cg': # TODO coingecko API
      self.logger.warning('Unable to update prices: CoinGecko API is not implemented yet.')
      return False
    else:
      self.logger.warning('Unable to update prices: "config.json" does not define a valid API provider.')
      return False
    dt = format_datetime(datetime.datetime.now())
    sql = []
    for slug,v in dic.items():
      if v.price is not None:
        sql.append(('UPDATE Token SET Price=?, LastUpdate=? WHERE Id=?;',(v.price,dt,v.ide)))
    res = self.db_execute(sql)
    if res is None:
      self.logger.warning('Unable to update prices: Database error.')
      return False
    self.last_update = datetime.datetime.now()
    return True
  def load_data(self) -> None:
    """
    Loads the entire database in the memory.
    Should be called after each database modification.
    """
    self.data = SimpleNamespace()
    array = if_none(self.db_execute(('SELECT * FROM Blockchain;',())),[])
    self.data.blockchain = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.name = row[1]
      record.ecosystem = row[2]
      record.is_ecosystem = False
      self.data.blockchain[record.ide] = record
    for blockchain in self.data.blockchain.values():
      if blockchain.ecosystem is not None:
        ecosystem = self.data.blockchain[blockchain.ecosystem]
        ecosystem.is_ecosystem = True
        blockchain.ecosystem = ecosystem
    array = if_none(self.db_execute(('SELECT * FROM CEX;',())),[])
    self.data.cex = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.name = row[1]
      self.data.cex[record.ide] = record
    array = if_none(self.db_execute(('SELECT * FROM Tag;',())),[])
    self.data.tag = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.name = row[1]
      self.data.tag[record.ide] = record
    array = if_none(self.db_execute(('SELECT * FROM Wallet;',())),[])
    self.data.wallet = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.name = row[1]
      record.cold = bool(row[2])
      self.data.wallet[record.ide] = record
    array = if_none(self.db_execute(('SELECT * FROM Token;',())),[])
    self.data.token = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.symbol = row[1]
      record.name = row[2]
      record.slug_cg = row[3]
      record.slug_cmc = row[4]
      record.underlying = row[5]
      record.decimals = row[6]
      record.price = row[7]
      record.slug_link = self.get_slug_link(record)
      record.last_update = datetime.datetime.strptime(row[8], DTFORMAT)
      self.data.token[record.ide] = record
    for token in self.data.token.values():
      if token.underlying is not None:
        underlying = self.data.token[token.underlying]
        underlying.is_underlying = True
        token.underlying = underlying
    array = if_none(self.db_execute(('SELECT * FROM Position;',())),[])
    self.data.position = {}
    for row in array:
      record = SimpleNamespace()
      record.ide = row[0]
      record.name = row[1]
      record.wallet = row[2]
      if record.wallet is not None:
        record.wallet = self.data.wallet[record.wallet]
      record.blockchain = row[3]
      if record.blockchain is not None:
        record.blockchain = self.data.blockchain[record.blockchain]
      record.cex = row[4]
      if record.cex is not None:
        record.cex = self.data.cex[record.cex]
      record.info = row[5]
      record.last_update = datetime.datetime.strptime(row[6], DTFORMAT)
      record.next_update = row[7]
      if record.next_update is not None:
        record.next_update = datetime.datetime.strptime(record.next_update, DTFORMAT)
      record.tags = []
      record.components = []
      self.data.position[record.ide] = record
    array = if_none(self.db_execute(('SELECT * FROM LinkTag;',())),[])
    for row in array:
      self.data.position[row[0]].tags.append(self.data.tag[row[1]])
    array = if_none(self.db_execute(('SELECT * FROM LinkToken;',())),[])
    for row in array:
      record = SimpleNamespace()
      record.token = self.data.token[row[1]]
      record.balance = row[2]
      record.apr = row[3]
      record.value = record.balance * record.token.price
      self.data.position[row[0]].components.append(record)
    for position in self.data.position.values():
      position.value = sum([component.value for component in position.components])
      position.apr = sum([component.value * component.apr for component in position.components]) / position.value if position.value != 0 else 0
    self.data.total_value = sum([position.value for position in self.data.position.values()])
  def get_slug_link(self, token:SimpleNamespace) -> Optional[str]:
    """
    Returns the CoinMarketCap or CoinGecko URL for a given token, or None
    Prioritize to the one defined in the config file.
    """
    if self.config['token_viewer'] == 'cmc' and token.slug_cmc is not None:
      return CMC_VIEWER + token.slug_cmc
    elif self.config['token_viewer'] == 'cg' and token.slug_cg is not None:
      return CG_VIEWER + token.slug_cg
    elif token.slug_cmc is not None:
      return CMC_VIEWER + token.slug_cmc
    elif token.slug_cg is not None:
      return CG_VIEWER + token.slug_cg
    return None
  def define_rules(self) -> None:
    """
    Defines valide routes for the web server and link them to function.
    Defines functions used to format data with jinja.
    """
    # filters jinja
    self.server.jinja_env.filters['if_none'] = if_none
    self.server.jinja_env.filters['fixed'] = fixed
    self.server.jinja_env.filters['format_price_round'] = format_price_round
    self.server.jinja_env.filters['format_price'] = format_price
    self.server.jinja_env.filters['format_datetime'] = format_datetime
    self.server.jinja_env.filters['format_time_remaining'] = format_time_remaining
    self.server.jinja_env.filters['format_time_elapsed'] = format_time_elapsed
    # index
    self.server.add_url_rule(rule='/', view_func=self.page_index)
    self.server.add_url_rule(rule='/ajax-index-update/', view_func=self.ajax_index_update, methods=['POST'])
    # update-prices
    self.server.add_url_rule(rule='/update-prices/', view_func=self.page_update_prices)
    self.server.add_url_rule(rule='/ajax-update-prices-before/', view_func=self.ajax_update_prices_before, methods=['POST'])
    self.server.add_url_rule(rule='/ajax-update-prices-after/', view_func=self.ajax_update_prices_after, methods=['POST'])
    # database
    self.server.add_url_rule(rule='/database/<table>', view_func=self.page_database)
    self.server.add_url_rule(rule='/ajax-database-delete/', view_func=self.ajax_database_delete, methods=['POST'])
    self.server.add_url_rule(rule='/form/', view_func=self.page_form, methods=['POST'])
    self.server.add_url_rule(rule='/recv-form/', view_func=self.recv_form, methods=['POST'])
    # view
    self.server.add_url_rule(rule='/view/', view_func=self.page_view, methods=['GET'])
    # assets
    self.server.add_url_rule(rule='/assets/', view_func=self.page_assets)
    # footer pages
    self.server.add_url_rule(rule='/log/', view_func=self.page_log)
    self.server.add_url_rule(rule='/contact/', view_func=self.page_contact)
    self.server.add_url_rule(rule='/doc/', view_func=self.page_doc)
    self.server.add_url_rule(rule='/donate/', view_func=self.page_donate)
    self.server.add_url_rule(rule='/license/', view_func=self.page_license)
  @staticmethod
  def page_error(title:str='', html:str='') -> Response:
    params = {
      'title':title,
      'html':html,
      'styles':['error']
    }
    return make_response(render_template('error.html', params=params))
  @staticmethod
  def page_error_database(sql:SQLRequests) -> Response:
    if isinstance(sql,list):
      title = 'Could not execute the following sequence of requests'
      html = '<ol>'+(''.join(list(map(lambda x: f'<li>{x[0]}<br>{x[1]}</li>',sql))))+'</ol>'
    else:
      title = 'Could not execute the following request'
      html = f'<p>{sql[0]}<br>{sql[1]}</p>'
    return Portfolio.page_error(title, html)
  @staticmethod
  def page_index() -> Response:
    params = {
      'scripts':['index'],
      'styles':['index']
    }
    return make_response(render_template('index.html', params=params))
  def ajax_index_update(self) -> Response:
    """
    Updates prices if delay (in seconds) is elapsed since last update, then returns portfolio value.
    If delay is not set, it means the auto update is disabled.
    """
    delay = int(request.json.get('delay'))
    if delay>=0 and (self.last_update is None or self.last_update+datetime.timedelta(seconds=delay)<=datetime.datetime.now()):
      if not self.update_prices():
        return Portfolio.page_error('Prices could not be updated')
      self.load_data()
    response = {
      'total_value': self.data.total_value,
      'last_update': format_datetime(self.last_update)
    }
    return jsonify(response)
  @staticmethod
  def page_update_prices() -> Response:
    params = {
      'scripts':['update_prices'],
      'styles':['update_prices']
    }
    return make_response(render_template('update_prices.html', params=params))
  def get_token_info(self) -> Dict:
    res = {}
    for token in self.data.token.values():
      res[token.ide] = {
        'ide':token.ide,
        'symbol':token.symbol,
        'name':token.name,
        'slug_link':token.slug_link,
        'price':token.price,
        'last_update':format_datetime(token.last_update)
      }
    return res
  def ajax_update_prices_before(self) -> Response:
    return jsonify({'tokens':self.get_token_info()})
  def ajax_update_prices_after(self) -> Response:
    tokens = self.get_token_info()
    if not self.update_prices():
      return Portfolio.page_error('Prices could not be updated')
    self.load_data()
    for idt,token in self.data.token.items():
      old_price = tokens[idt]['price']
      tokens[idt]['new_price'] = token.price
      tokens[idt]['variation'] = (token.price-old_price)/old_price*100 if old_price!=0 else 0
      if format_datetime(token.last_update)!=tokens[idt]['last_update'] or token.price!=old_price:
        tokens[idt]['last_update'] = None # marks that the update is successful
    response = {
      'tokens':tokens,
      'small_variation':self.config['percent_small_variation'],
      'big_variation':self.config['percent_big_variation']
    }
    return jsonify(response)
  def page_database(self, table:str) -> Response:
    if table not in ['Position','Token','Blockchain','Wallet','CEX','Tag']:
      return redirect('/database/Position')
    params = {
      'scripts':['database'],
      'styles':['database'],
      'table':table,
      'data':self.data
    }
    return make_response(render_template('database.html', params=params))
  def ajax_database_delete(self) -> Response:
    form = request.get_json()
    table, ide = form['table'], form['ide']
    sql = []
    if table=='Tag':
      sql.append(('DELETE FROM LinkTag WHERE IdTag=?',(ide,)))
    elif table=='Position':
      sql.append(('DELETE FROM LinkTag WHERE IdPosition=?;',(ide,)))
      sql.append(('DELETE FROM LinkToken WHERE IdPosition=?;',(ide,)))
    sql.append((f'DELETE FROM {table} WHERE Id=?;',(ide,)))
    if self.db_execute(sql) is None:
      self.logger.warning('Error while deleting in the database.')
      return Portfolio.page_error_database(sql)
    self.load_data()
    return jsonify({})
  def page_form(self) -> Response:
    table, origin = request.form['table'], request.form['origin']
    ide = int(request.form['ide']) if request.form['ide']!='' else ''
    params = {
      'styles':['form'],
      'scripts':['form_position'] if table=='Position' else [],
      'data':self.data,
      'origin':origin,
      'ide':ide,
      'table':table
    }
    return make_response(render_template('form.html', params=params))
  def recv_form(self) -> Response:
    form = request.form
    table, ide = form['table'], form['ide']
    sql = []
    if ide!='' and table=='Position':
      sql.append(('DELETE FROM LinkTag WHERE IdPosition=?;',(ide,)))
      sql.append(('DELETE FROM LinkToken WHERE IdPosition=?;',(ide,)))
    if table=='Tag' or table=='CEX':
      name = form['name']
      values = (name,)
      if ide=='':
        sql.append((f'INSERT INTO {table} VALUES (null,?);',values))
      else:
        sql.append((f'UPDATE {table} SET Name=? WHERE Id=?;',values+(ide,)))
    elif table=='Wallet':
      name = form['name']
      cold = 1 if 'cold' in form else 0
      values = (name,cold)
      if ide=='':
        sql.append(('INSERT INTO Wallet VALUES (null,?,?);',values))
      else:
        sql.append(('UPDATE Wallet SET Name=?, Cold=? WHERE Id=?;',values+(ide,)))
    elif table=='Blockchain':
      name = form['name']
      ecosystem = str_or_none(form['ecosystem'])
      values = (name,ecosystem)
      if ide == '':
        sql.append(('INSERT INTO Blockchain VALUES (null,?,?);',values))
      else:
        sql.append(('UPDATE Blockchain SET Name=?, Ecosystem=? WHERE Id=?;',values+(ide,)))
    elif table=='Token':
      name = form['name']
      symbol = form['symbol']
      underlying = str_or_none(form['underlying'])
      slug_cmc = str_or_none(form['slug_cmc'])
      slug_cg = str_or_none(form['slug_cg'])
      decimals = form['decimals']
      price = form['price']
      last_update = format_datetime(datetime.datetime.now())
      values = (symbol,name,slug_cg,slug_cmc,underlying,decimals,price,last_update)
      if ide == '':
        sql.append(('INSERT INTO Token VALUES (null,?,?,?,?,?,?,?,?);',values))
      else:
        sql.append(('UPDATE Token SET Symbol=?, Name=?, SlugCG=?, SlugCMC=?, Underlying=?, Decimals=?, Price=?, LastUpdate=? WHERE Id=?;',values+(ide,)))
    else:
      now = datetime.datetime.now()
      name = form['name']
      info = form['info']
      last_update = now.strftime(DTFORMAT)
      if form['custody_type']=='0':
        wallet = str_or_none(form['wallet'])
        blockchain = str_or_none(form['blockchain'])
        cex = None
      else:
        wallet = None
        blockchain = None
        cex = str_or_none(form['cex'])
      if form['next_update_type']=='0':
        next_update = now+datetime.timedelta(days=int(form['next_update_reference'])) if form['next_update_reference']!='' else None
      else:
        next_update = datetime.datetime.strptime(form['next_update_custom'].replace('T',' '),DTFORMAT) if form['next_update_custom']!='' else None
      if next_update is not None:
        next_update = format_datetime(next_update)
      values = (name,wallet,blockchain,cex,info,last_update,next_update)
      if ide=='':
        sql.append(('INSERT INTO Position VALUES (null,?,?,?,?,?,?,?);',values))
        sql.append(('SELECT seq FROM sqlite_sequence WHERE name="Position";',()))
        res = self.db_execute(sql)
        if res is None:
          self.logger.warning('Error while writing in the database.')
          return Portfolio.page_error_database(sql)
        else:
          ide = res[0][0]
          sql = []
      else:
        sql.append(('UPDATE Position SET Name=?, Wallet=?, Blockchain=?, CEX=?, Info=?, LastUpdate=?, NextUpdate=? WHERE Id=?;',values+(ide,)))
      tags = form['tags'].split('-') if form['tags']!='' else []
      if len(tags)>0:
        values = []
        for tag in tags:
          values.extend((ide,tag))
        question_marks = ','.join(['(?,?)']*len(tags))
        sql.append((f'INSERT INTO LinkTag VALUES {question_marks};',tuple(values)))
      values = []
      count = 0
      for field in form.keys():
        if field[:6]=='token-':
          count += 1
          n = field[6]
          values.extend((ide,form['token-'+n],form['balance-'+n],form['apr-'+n]))
      if count>0:
        question_marks = ','.join(['(?,?,?,?)']*count)
        sql.append((f'INSERT INTO LinkToken VALUES {question_marks};',tuple(values)))
    if self.db_execute(sql) is None:
      self.logger.warning('Error while writing in the database.')
      return Portfolio.page_error_database(sql)
    self.load_data()
    return redirect(request.form['origin'])
  def page_view(self) -> Response:
    filter_type = request.args.get('filter-type')
    filter_value = request.args.get('filter-value')
    group = request.args.get('group')
    order = request.args.get('order')
    if any([v is None for v in [filter_type, filter_value, group, order]]):
      filter_type,filter_value,group,order = 'all','all','nothing','value'
    now = datetime.datetime.now()
    groups = {}
    try:
      for position in self.data.position.values():
        special_value = None
        keep = False
        if filter_type=='all':
          keep = True
        elif filter_type=='wallet':
          keep = position.wallet is not None and position.wallet.name==filter_value
        elif filter_type=='blockchain':
          keep = position.blockchain is not None and position.blockchain.name==filter_value
        elif filter_type=='cex':
          keep = position.cex is not None and position.cex.name==filter_value
        elif filter_type=='ecosystem':
          keep = (position.blockchain is not None) and ((position.blockchain.name==filter_value) or (position.blockchain.ecosystem is not None and position.blockchain.ecosystem.name==filter_value))
        elif filter_type=='underlying':
          special_value = 0
          for component in position.components:
            if (component.token.symbol==filter_value) or (component.token.underlying is not None and component.token.underlying.symbol==filter_value):
              keep = True
              special_value += component.balance*component.token.price
        elif filter_type=='token':
          special_value = 0
          for component in position.components:
            if component.token.symbol==filter_value:
              keep = True
              special_value += component.balance * component.token.price
        elif filter_type=='tag':
          keep = filter_value in [tag.name for tag in position.tags]
        elif filter_type=='custody':
          if filter_value=='self':
            keep = position.wallet is not None
          elif filter_value=='cold':
            keep = position.wallet is not None and position.wallet.cold
          elif filter_value=='hot':
            keep = position.wallet is not None and not position.wallet.cold
          elif filter_value=='third-party':
            keep = position.wallet is None
          else:
            raise ValueError(f'Custody "{filter_value}" is not valid.')
        elif filter_type=='update':
          if position.next_update is None:
            keep = False
          else:
            if filter_value=='now':
              keep = position.next_update<=now
            elif filter_value=='1d':
              keep = position.next_update<=now+datetime.timedelta(days=1)
            elif filter_value=='1w':
              keep = position.next_update<=now+datetime.timedelta(weeks=1)
            else:
              raise ValueError(f'Update "{filter_value}" is not valid.')
        else:
          raise ValueError(f'Filter "{filter_type} + {filter_value}" is not valid.')
        if keep:
          position.tags.sort(key=lambda tag: tag.name)
          position.components.sort(key=lambda component: (component.token.symbol,-component.token.price))
          priority = 0 #prevails over the "order" criteria
          if group=='nothing':
            group_name = position.name
          elif group=='wallet':
            if position.wallet is not None:
              group_name = position.wallet.name
            else:
              group_name = 'Other : Third-Party'
              priority = -1
          elif group=='blockchain':
            if position.blockchain is not None:
              group_name = position.blockchain.name
            else:
              group_name = 'Other : Third-Party'
              priority = -1
          elif group=='cex':
            if position.cex is not None:
              group_name = position.cex.name
            else:
              group_name = 'Other : Self-Custody'
              priority = -1
          elif group=='ecosystem':
            if position.blockchain is not None:
              if position.blockchain.ecosystem is not None:
                group_name = position.blockchain.ecosystem.name
              elif position.blockchain.is_ecosystem:
                group_name = position.blockchain.name
              else:
                group_name = 'Other : No ecosystem'
                priority = -1
            else:
              group_name = 'Other : No ecosystem'
              priority = -1
          elif group=='custody':
            if position.wallet is None:
              group_name = 'Third-party'
              priority = 1
            elif position.wallet.cold:
              group_name = 'Cold'
              priority = 3
            else:
              group_name = 'Hot'
              priority = 2
          elif group=='update':
            if position.next_update is None:
              group_name = 'Other : Undefined'
              priority = -1
            elif position.next_update<=now:
              group_name = 'Now'
              priority = 4
            elif position.next_update<=now+datetime.timedelta(days=1):
              group_name = 'Less than 1 day'
              priority = 3
            elif position.next_update<=now+datetime.timedelta(weeks=1):
              group_name = 'Less than 1 week'
              priority = 2
            else:
              group_name = 'More than 1 week'
              priority = 1
          else:
            raise ValueError(f'Group "{group}" is not valid.')
          if not group_name in groups.keys():
            gr = SimpleNamespace()
            gr.name = group_name
            gr.positions = [position]
            gr.priority = priority
            if special_value is None:
              gr.value = position.value
              gr.apr = 0
            else:
              gr.value = special_value
              gr.apr = None
            groups[group_name] = gr
          else:
            groups[group_name].positions.append(position)
            if special_value is None:
              groups[group_name].value += position.value
            else:
              groups[group_name].value += special_value
      groups = list(groups.values())
      for gr in groups:
        if gr.apr is not None:
          gr.apr = sum([position.value*position.apr for position in gr.positions])/gr.value if gr.value!=0 else 0
        gr.next_update = min([position.next_update for position in gr.positions if position.next_update is not None], default=None)
        gr.last_update = min([position.last_update for position in gr.positions])
        gr.positions.sort(key=lambda position: position.value, reverse=True)
      selection_value = sum([gr.value for gr in groups])
      selection_apr = sum([gr.value*gr.apr for gr in groups])/selection_value if selection_value!=0 and groups[0].apr is not None else 0
      daily_earnings = selection_value*selection_apr/100/365.25
      monthly_earnings = selection_value*selection_apr/100/12
      groups.sort(key=lambda gr: abs(gr.value), reverse=True)
      chart = []
      shuffle(COLORS)
      other_value = 0
      j = 0
      for i,gr in enumerate(groups):
        if i >= len(COLORS):
          other_value += gr.value
        else:
          part = SimpleNamespace()
          part.name = gr.name
          part.value = gr.value
          if gr.value>0:
            part.color = COLORS[i]
          else:
            part.color = COLORS_DEBT[j]
            j += 1
          chart.append(part)
      if other_value!=0:
        part = SimpleNamespace()
        part.name = 'OTHER'
        part.value = other_value
        part.color = '#FFFFFF' if other_value>0 else '#000000'
        chart.append(part)
      if order=='value':
        groups.sort(key=lambda gr: (-gr.priority,-gr.value))
      elif order=='alpha':
        groups.sort(key=lambda gr: (-gr.priority,gr.name))
      elif order=='last-update':
        groups.sort(key=lambda gr: (-gr.priority,gr.last_update))
      elif order=='next-update':
        groups.sort(key=lambda gr: (-gr.priority,if_none(gr.next_update,datetime.datetime.max)))
      else:
        raise ValueError(f'Order "{order}" is not valid.')
    except ValueError as e:
      return self.page_error(html=f'<p>{str(e)}</p>')
    params = {
      'scripts':['chart','view'],
      'styles':['view'],
      'filter_type':filter_type,
      'filter_value':filter_value,
      'group':group,
      'order':order,
      'data':self.data,
      'groups':groups,
      'selection_value':selection_value,
      'selection_apr':selection_apr,
      'daily_earnings':daily_earnings,
      'monthly_earnings':monthly_earnings,
      'chart':chart
    }
    return make_response(render_template('view.html', params=params))
  @staticmethod
  def get_asset(assets:dict, token:SimpleNamespace) -> SimpleNamespace:
    if token.ide in assets:
      return assets[token.ide]
    asset = SimpleNamespace()
    asset.ide = token.ide
    asset.symbol = token.symbol
    asset.slug_link = token.slug_link
    asset.price = token.price
    asset.decimals = token.decimals
    asset.balance = 0
    asset.value = 0
    asset.derivative = False
    assets[asset.ide] = asset
    return asset
  def page_assets(self) -> Response:
    assets = {}
    for position in self.data.position.values():
      for component in position.components:
        asset = self.get_asset(assets, component.token)
        asset.balance += component.balance
        asset.value = asset.price*asset.balance
        if component.token.underlying is not None:
          asset.derivative = True
          underlying = self.get_asset(assets, component.token.underlying)
          if underlying.price>0:
            underlying.value += asset.price*component.balance
            underlying.balance += asset.price/underlying.price*component.balance
    assets = list(assets.values())
    assets.sort(key=lambda asset: (asset.derivative,-abs(asset.value)))
    chart = []
    shuffle(COLORS)
    other_value = 0
    for i, asset in enumerate(assets):
      if asset.derivative:
        break
      if i>=len(COLORS):
        other_value += asset.value
      else:
        part = SimpleNamespace()
        part.name = asset.symbol
        part.value = asset.value
        part.color = COLORS[i]
        chart.append(part)
    if other_value>0:
      part = SimpleNamespace()
      part.name = 'OTHER'
      part.value = other_value
      part.color = '#666666'
      chart.append(part)
    params = {
      'styles':['database','assets'],
      'scripts':['chart','assets'],
      'assets':assets,
      'chart':chart,
      'total_value':self.data.total_value
    }
    return make_response(render_template('assets.html', params=params))
  @staticmethod
  def page_log() -> Response:
    return send_file('portfolio.log', mimetype='text/plain')
  @staticmethod
  def page_contact() -> Response:
    return make_response(render_template('contact.html'))
  @staticmethod
  def page_donate() -> Response:
    params = {
      'styles':['donate'],
      'scripts':['donate'],
    }
    return make_response(render_template('donate.html', params=params))
  @staticmethod
  def page_license() -> Response:
    return send_file('license.txt', mimetype='text/plain')
  @staticmethod
  def page_doc() -> Response:
    return make_response(render_template('doc.html'));

if __name__=='__main__':
  Portfolio().run()