var FilterType;
var FilterValue;
var Group;
var Order;
var ChartLabels = [];
var ChartValues = [];
var ChartColors = [];

$(document).ready(Init);

function Init(){
  GetGlobals();
  InitFilterType();
  InitFilterValue();
  InitGroup();
  InitOrder();
  $('#filter-type').on('change',FilterTypeChange);
  $('#filter-value').on('change',FilterValueChange);
  DrawChart();
  $('img[data-action="edit"]').each(function(){$(this).on('click',function(){ActionEdit($(this).closest('div'));});});
  $('img[data-action="delete"]').each(function(){$(this).on('click',function(){ActionDelete($(this).closest('div'));});});
  $('img[data-action="info"]').each(function(){$(this).on('click',function(){$(this).closest('section.position').find('.info-container').toggleClass('not-displayed');});});
}

function ActionEdit(div){
  $('#ide').val($(div).data('ide'));
  $('#edit-form').submit();
}

function ActionDelete(div){
  let confirmation = confirm('Delete ?');
  if(!confirmation) return;
  $.ajax({
    url: '/ajax-database-delete',
    data: JSON.stringify({table: 'Position', ide: $(div).data('ide')}),
    success: function(response){
      Refresh();
    }
  });
}

function LoadChart(){
  $('.chart-part').each(function(){
    ChartLabels.push($(this).data('label'));
    ChartValues.push($(this).data('value'));
    ChartColors.push($(this).data('color'));
  })
}

function DrawChart(){
  LoadChart();
  if(ChartLabels.length==0){return;}
  var ctx = $('#chart')[0].getContext('2d');
  var chart = new Chart(ctx,{
    type:'doughnut',
    data:{
      labels:ChartLabels,
      datasets:[{
        data:ChartValues,
        backgroundColor:ChartColors
      }]
    },
    options: {
      responsive:false,
      plugins:{
        legend:{
          position:'right',
          align:'middle',
          labels:{
            color:'#00296B',
            font:{
              size:16
            }
          }
        },
        tooltip:{
          callbacks:{
            label: ({ label, formattedValue }) => FormatPriceRound(parseInt(formattedValue.replace(/\s/g,'')))+' $'
          }
        }
      },
      elements:{
        arc:{
          borderColor:'#00296B',
        }
      }
    }
  });
}

function GetGlobals(){
  FilterType = $('#get-filter-type').data('get');
  FilterValue = $('#get-filter-value').data('get');
  Group = $('#get-group').data('get');
  Order = $('#get-order').data('get');
}

function InitFilterType(){
  $('#filter-type').val(FilterType);
}

function FilterTypeChange(){
  FilterType = $('#filter-type').val();
  InitFilterValue();
  InitGroup();
}

function InitFilterValue(){
  $('#filter-value').html($('#filter-value-'+FilterType).html());
  $('#filter-value').val(FilterValue);
  if($('#filter-value').val()==null){
    $('#filter-value option:first').prop('selected',true);
  }
}

function FilterValueChange(){
  FilterValue = $('#filter-value').val();
  InitGroup();
}

function InitGroup(){
  let lgroups = ['Nothing'];
  if(['all','tag','token','underlying'].includes(FilterType) || (FilterType=='update' && FilterValue!='now')){
    lgroups = ['Nothing','Wallet','Ecosystem','Blockchain','Custody','CEX','Update'];
  }
  else if(FilterType=='update' && FilterValue=='now'){
    lgroups = ['Nothing','Wallet','Ecosystem','Blockchain','Custody','CEX'];
  }
  else if(FilterType=='custody'){
    if(FilterValue=='cold' || FilterValue=='hot'){
      lgroups = ['Nothing','Wallet','Ecosystem','Blockchain','Update'];
    }
    else if(FilterValue=='self'){
      lgroups = ['Nothing','Wallet','Ecosystem','Blockchain','Custody','Update'];
    }
    else if(FilterValue=='third-party'){
      lgroups = ['Nothing','CEX','Update'];
    }
  }
  else if(FilterType=='cex'){
    lgroups = ['Nothing','Update'];
  }
  else if(FilterType=='wallet'){
    lgroups = ['Nothing','Ecosystem','Blockchain','Update'];
  }
  else if(FilterType=='blockchain'){
    lgroups = ['Nothing','Wallet','Custody','Update'];
  }
  else if(FilterType=='ecosystem'){
    lgroups = ['Nothing','Wallet','Blockchain','Custody','Update'];
  }
  let options = '';
  for(let elem of lgroups.values()){
    let Elem = elem;
    elem = Elem.toLowerCase();
    options += '<option value="'+elem+'">'+Elem+'</option>';
  }
  $('#group').html(options);
  $('#group').val(Group);
  if($('#group').val()==null){
    $('#group option:first').prop('selected',true);
  }
}

function InitOrder(){
  $('input[name="order"][value="'+Order+'"]').prop('checked',true);
}